@@ -459,6 +459,247 @@ gamm40 <- function(formula,random=NULL,family=gaussian(),data=list(),weights=NUL
 
 } ## end of gamm40
 
+
+gamm4.oldwork <- function(G,mf,n.sr,r.name,family,formula,gam.terms,pTerms,lme4.formula,linear) {
+## does the model fitting work if an old version of lme4 is loaded
+  if (linear) b <- lmer(lme4.formula,data=mf,family=family,weights=G$w,doFit=FALSE)
+  else  b <- glmer(lme4.formula,data=mf,family=family,weights=G$w,doFit=FALSE)
+
+  if (n.sr) { ## use Fabian Scheipl's trick of overwriting dummy slots
+     tn <- names(b$FL$fl) 
+     ## some names go with more than one element of b$FL$trms, so...
+     tn <- tn[attr(b$FL$fl,"assign")] ## group name associated with each element of b$FL$trms 
+     ind <- 1:length(tn)
+     sn <- names(G$random)
+     for (i in 1:n.sr) { ## loop through random effect smooths
+       k <- ind[sn[i]==tn] ## which trm should contain G$random[[i]] 
+       b$FL$trms[[k]]$A <- b$FL$trms[[k]]$Zt <- as(t(G$random[[i]]),"dgCMatrix")
+       attr(G$random[[i]],"s.label") -> sl
+       attr(b$FL$trms[[k]]$ST,"dimnames") <- list(sl,sl)
+     }
+  }
+
+  ret <- list()
+
+  if (linear) ret$mer <- do.call(lme4:::lmer_finalize,b)
+  else ret$mer <- do.call(lme4:::glmer_finalize,b)
+
+  rm(b)
+
+  ### .... fitting finished
+
+  ## now fake a gam object 
+    
+  object<-list(model=mf,formula=formula,smooth=G$smooth,nsdf=G$nsdf,family=family,
+                 df.null=nrow(G$X),y=ret$mer@y, ## getME problem - don't know how else to extract
+                 terms=gam.terms,pterms=pTerms,xlevels=G$xlevels,
+                 contrasts=G$contrasts,assign=G$assign,na.action=attr(mf,"na.action"),
+                 cmX=G$cmX,var.summary=G$var.summary)
+  
+  ## to unpack coefficients look at names(ret$lme$flist), ret$lme@Zt, ranef(), fixef()
+ 
+    ## let the GAM coefficients in the original parameterization be beta,
+    ## and let them be beta' in the fitting parameterization. 
+    ## Then beta = B beta'. B and B^{-1} can be efficiently accumulated
+    ## and are useful for stable computation of the covariance matrix
+    ## etc... 
+  
+    B <- Matrix(0,ncol(G$Xf),ncol(G$Xf))
+    diag(B) <- 1
+    Xfp <- G$Xf
+    ## Transform  parameters back to the original space....
+    bf <- as.numeric(lme4::fixef(ret$mer)) ## the fixed effects
+    br <- lme4::ranef(ret$mer) ## a named list
+    if (G$nsdf) p <- bf[1:G$nsdf] else p <- array(0,0) ## fixed parametric componet
+    if (G$m>0) for (i in 1:G$m) {
+      fx <- G$smooth[[i]]$fixed 
+      first <- G$smooth[[i]]$first.f.para; last <- G$smooth[[i]]$last.f.para
+      if (first <=last) beta <- bf[first:last] else beta <- array(0,0)
+      if (fx) b <- beta else { ## not fixed so need to undo transform of random effects etc. 
+        b <- rep(0,0)
+        for (k in 1:length(G$smooth[[i]]$lmer.name)) ## collect all coefs associated with this smooth
+          b <- c(b,as.numeric(br[[G$smooth[[i]]$lmer.name[k]]][[1]]))     
+        b <- b[G$smooth[[i]]$rind] ## make sure coefs are in order expected by smooth
+        b <- c(b,beta) 
+        b <- G$smooth[[i]]$trans.D*b
+        if (!is.null(G$smooth[[i]]$trans.U)) b <- G$smooth[[i]]$trans.U%*%b ## transform back to original 
+      }
+      p <- c(p,b)
+     
+      ## now fill in B...
+      ind <- G$smooth[[i]]$first.para:G$smooth[[i]]$last.para
+      if (!fx) { 
+         D <- G$smooth[[i]]$trans.D
+         if (is.null(G$smooth[[i]]$trans.U)) B[ind,ind] <- Diagonal(length(D),D) else
+         B[ind,ind] <- t(D*t(G$smooth[[i]]$trans.U))
+      }
+      ## and finally transform G$Xf into fitting parameterization...
+      Xfp[,ind] <- G$Xf[,ind,drop=FALSE]%*%B[ind,ind,drop=FALSE]
+
+    }
+ 
+    object$coefficients <- p
+
+    ## need to drop smooths from Zt and then
+    ## form Z'phiZ + I \sigma^2
+
+    vr <- lme4::VarCorr(ret$mer) ## list of ranef variance components in the same order as Gp
+    
+    scale <- as.numeric(attr(vr,"sc"))^2 ## get the scale parameter
+    if (!is.finite(scale)) { 
+      scale <- 1
+      object$scale.estimated <- FALSE
+    } else object$scale.estimated <- TRUE
+    
+    ## The relative variance matrices for each random effect are stored in factorized form TSST' 
+    ## where T is lower tri and S diagonal. `expand' will extract the same factorization 
+    ## for the whole of phi, using sparse matrices. expand returns a permutation matrix P
+    ## but at time of writing the ordering of T and S and Z are consistent without 
+    ## permutation, and inconsistent with it, so it appears to be irrelevant here. 
+
+    phi <- Matrix::expand(ret$mer,sparse = TRUE) ## factorization of phi as TSST'
+        
+    sp <- rep(-1,n.sr)
+
+    Zt <- Matrix(0,0,ncol(getME(ret$mer,"Zt")))
+    if (n.sr==0) sn <- NULL ## names by which smooths are known in mer
+    rn <- names(vr)
+    ind <- rep(0,0) ## index the non-smooth random effects among the random effects
+    for (i in 1:length(vr)) {
+      if (is.null(sn)||!rn[i]%in%sn) { ## append non smooth r.e.s to Zt
+        Gp <- getME(ret$mer,"Gp")
+        ind <- c(ind,(Gp[i]+1):Gp[i+1])
+      } else if (!is.null(sn)) { ## extract smoothing parameters for smooth r.e.s
+        k <- (1:n.sr)[rn[i]==sn] ## where in original smooth ordering is current smoothing param
+        if (as.numeric(vr[[i]]>0)) sp[k] <- scale/as.numeric(vr[[i]]) else 
+        sp[k] <- 1e10
+      }
+    }
+
+    if (length(ind)) { ## extract columns corresponding to non-smooth r.e.s 
+      ## extracting random effects model matrix
+      Zt <- getME(ret$mer,"Zt")[ind,]
+      root.phi <- phi$S[ind,ind]%*%t(phi$T[ind,ind]) ## and corresponding sqrt of cov matrix (phi)
+    }
+
+    object$prior.weights <- ret$mer@pWt ## getME problem --- not accessible
+
+    if (length(ret$mer@var)==0) { 
+      object$weights <- object$prior.weights
+      V <- Diagonal(n=length(object$weights),x=scale/object$weights) 
+    } else 
+    {  ## ret$mer@sqrtrWt is root reciprocal weights...
+       V <- Diagonal(x=ret$mer@sqrtrWt^2)*scale ## the response variance conditional on the r.e.s
+       object$weights <- 1/ret$mer@sqrtrWt^2
+       #object$weights <- 1/ret$mer@var 
+    }
+  
+    if (nrow(Zt)>0) V <- V + crossprod(root.phi%*%Zt)*scale ## data or pseudodata cov matrix, treating smooths as fixed now
+
+    ## NOTE: Cholesky probably better in the following - then pivoting 
+    ##       automatic when solving....
+
+    R <- Matrix:::chol(V,pivot=TRUE);piv <- attr(R,"pivot") 
+
+    G$Xf <- as(G$Xf,"dgCMatrix")
+    Xfp <- as(Xfp,"dgCMatrix")
+    
+    if (is.null(piv)) {
+      WX <- as(solve(t(R),Xfp),"matrix")    ## V^{-.5}Xp -- fit parameterization
+      XVX <- as(solve(t(R),G$Xf),"matrix")  ## same in original parameterization 
+    } else {
+      WX <- as(solve(t(R),Xfp[piv,]),"matrix")    ## V^{-.5}Xp -- fit parameterization
+      XVX <- as(solve(t(R),G$Xf[piv,]),"matrix")  ## same in original parameterization
+    }
+    qrz <- qr(XVX,LAPACK=TRUE)
+    object$R <- qr.R(qrz);object$R[,qrz$pivot] <- object$R
+
+    XVX <- crossprod(object$R) ## X'V^{-1}X original parameterization
+    object$sp <- sp
+    
+    colx <- ncol(G$Xf)
+    Sp <- matrix(0,colx,colx) # penalty matrix - fit param
+    first <- G$nsdf+1
+    k <- 1
+    if (G$m>0) for (i in 1:G$m) { # Accumulate the total penalty matrix
+      if (!object$smooth[[i]]$fixed) {
+        ii <- object$smooth[[i]]$first.para:object$smooth[[i]]$last.para ## index this smooth's params
+        for (j in 1:length(object$smooth[[i]]$S)) { ## work through penalty list
+          ind <- ii[object$smooth[[i]]$pen.ind == j] ## index of currently penalized
+          diag(Sp)[ind] <-  sqrt(object$sp[k]) ## diagonal penalty
+          k <- k+1
+        }														              }
+      first <- last + 1 
+    }
+   
+    ## Alternative cov matrix calculation. Basic
+    ## idea is that cov matrix is computed stably in
+    ## fitting parameterization, and then transformed to
+    ## original parameterization. 
+    qrx <- qr(rbind(WX,Sp/sqrt(scale)),LAPACK=TRUE)
+    Ri <- backsolve(qr.R(qrx),diag(ncol(WX)))
+    ind <- qrx$pivot;ind[ind] <- 1:length(ind)## qrx$pivot
+    Ri <- Ri[ind,] ## unpivoted square root of cov matrix in fitting parameterization Ri Ri' = cov
+    Vb <- B%*%Ri; Vb <- Vb%*%t(Vb)
+
+    object$edf<-rowSums(Vb*t(XVX))
+   
+    object$df.residual <- length(object$y) - sum(object$edf)
+
+    object$sig2 <- scale
+    if (linear) { object$method <- "lmer.REML"
+    } else { object$method <- "glmer.ML"}
+
+    object$Vp <- as(Vb,"matrix")
+  
+    object$Ve <- as(Vb%*%XVX%*%Vb,"matrix")
+   
+    class(object) <- "gam"
+   
+    ## Restore original smooth list, if it was split to deal with t2 terms...
+    if (!is.null(G$original.smooth)) {
+      object$smooth <- G$smooth <- G$original.smooth
+    }
+
+    ## If prediction parameterization differs from fit parameterization, transform now...
+    ## (important for t2 smooths, where fit constraint is not good for component wise 
+    ##  prediction s.e.s)
+
+    if (!is.null(G$P)) {
+      object$coefficients <- G$P %*% object$coefficients
+      object$Vp <- G$P %*% object$Vp %*% t(G$P) 
+      object$Ve <- G$P %*% object$Ve %*% t(G$P) 
+    }
+
+    object$linear.predictors <- predict.gam(object,type="link")
+    object$fitted.values <- object$family$linkinv(object$linear.predictors)
+    
+    object$residuals <- lme4:::residuals(ret$mer) 
+
+    if (G$nsdf>0) term.names<-colnames(G$X)[1:G$nsdf] else term.names<-array("",0)
+    n.smooth<-length(G$smooth) 
+    if (n.smooth)
+    for (i in 1:n.smooth)
+    { k<-1
+      for (j in object$smooth[[i]]$first.para:object$smooth[[i]]$last.para)
+      { term.names[j]<-paste(object$smooth[[i]]$label,".",as.character(k),sep="")
+        k<-k+1
+      }
+    }
+    names(object$coefficients) <- term.names  # note - won't work on matrices!!
+    names(object$edf) <- term.names
+    names(object$sp) <- names(G$sp)
+
+    object$gcv.ubre <- deviance(ret$mer)
+
+    if (!is.null(G$Xcentre)) object$Xcentre <- G$Xcentre ## any column centering applied to smooths
+
+    ret$gam<-object
+    class(gamm4) <- c("gamm4","list")
+    ret
+} ## gamm4.oldwork
+
+
 gamm4 <- function(formula,random=NULL,family=gaussian(),data=list(),weights=NULL,
       subset=NULL,na.action,knots=NULL,...) {
 # Routine to fit a GAMM to some data. Fixed and smooth terms are defined in the formula, but the wiggly 
@@ -466,11 +707,12 @@ gamm4 <- function(formula,random=NULL,family=gaussian(),data=list(),weights=NULL
 # random terms. 
 ## THIS VERSION is for `new' lme4
   if (packageVersion("lme4")<package_version("0.999999-999")) {
+    old.lme4 <- TRUE
     ## then simply call old version of gamm4.
-    mc <- match.call()
-    mc[[1]] <- as.name("gamm40")
-    return(eval(mc))
-  }
+#    mc <- match.call()
+#    mc[[1]] <- as.name("gamm40")
+#    return(eval(mc))
+  } else old.lme4 <- FALSE
 
   if (!require("lme4")) stop("gamm4() requires package lme4 to be installed")
   if (!require("mgcv")) stop("gamm4() requires package mgcv to be installed")
@@ -567,6 +809,8 @@ gamm4 <- function(formula,random=NULL,family=gaussian(),data=list(),weights=NULL
   
   lme4.formula <- as.formula(lme4.formula)
   
+  if (old.lme4) return(gamm4.oldwork(G,mf,n.sr,r.name,family,formula,gam.terms,pTerms,lme4.formula,linear))
+
   ## NOTE: further arguments should be passed here... 
   b <- if (linear) lFormula(lme4.formula,data=mf,weights=G$w,...) else 
                    glFormula(lme4.formula,data=mf,family=family,weights=G$w,...)
